
/***************************************************************************
 *  SimplIntercon Simulator Version 1.01
 *  Copyright (C) 2006 Hashem H. Najaf-abadi
 *  
 *  SimplIntercon is an adaptive wormhole-routing interconnection-network
 *  simulator developed by Hashem H. Najaf-abadi at IPM School of Computer 
 *  Science (Tehran - Iran), and North Carolina State University. 
 *
 *  This simulator is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License version
 *  2.1 published by the Free Software Foundation.
 *  
 *  Please check the Copyright and License information.
 *  
****************************************************************************/
 
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream.h>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <math.h>

char* output;

char adaptive; 
char wrap_around; 
int N_dimensns;
int L_dimensns;
int N_nodes;
int N_channels;
double gen_period;
double stop_period;
double period_step;
#define MAX_Time 4000000
#define BATCH_SIZE 10000
int MESSAGE_LENGTH;

#define NO_CYCLES 100   // Ratio of cycle time to tail propagation delay

#define HND_SHK 1
#define INJECT 2
#define ROUT_HEADER 3

///////////////////////////////////////////////////////////////////////////////////////////

int GEN_PRD;
int batches_counted;
FILE   *outp_file;

///////////////////////////////////////////////////////////////////////////////////////////

long int global, global1;
char turning_global;

class v_channels;
class p_channels;
class network;
class Time;
Time* main_Time;

class buffer {
private:
	int* destination_address;
	int length;
	int physical_c;
	int virtual_c;
	int* source;
	int frwd_bkwd;
	v_channels* from_channel;
	unsigned long int* insrtn_time_stamp;
	buffer** waiting_c;
public:
	char deterministic;
	char definit;
	int possible_ptr;
	buffer** possible;
	buffer* next;
	buffer* last_buf;
	p_channels* channel_ahead;
	char last_new_data();
	int* get_source() {
		return source;
	}
	void adjust_ptrs() {
		if (next != NULL)
			next->last_buf = last_buf;
		if (last_buf != NULL)
			last_buf->next = next;
		else *waiting_c = next;
	}
	int get_dst_frwd_bhnd() {
		return frwd_bkwd;
	}
	int from_pc() {
		return physical_c;
	}
	int from_vc() {
		return virtual_c;
	}
	char dec_length() {
		length--;
		if (length) return 1;
		else return 0;
	}
	int get_length() {
		return length;
	}
	int* get_destination_address() {
		return destination_address;
	}
	void del_destination_address() {
		delete destination_address;
	}
	unsigned long int* get_insrtn_time_stamp() {
		return insrtn_time_stamp;
	}

	buffer(int*, unsigned long int*, int, int, int, int*, int, v_channels*, p_channels*, 
	        buffer** in_possible, int in_possible_ptr, buffer* in_last_buf, char in_definit, 
	        buffer** in_waiting, char in_deterministic);

};

class v_channels {
private:
	char NewData_Recieved;
	buffer* input_buffer;
public:
	v_channels();
	char alocated();
	void aloct(buffer* input_buffer);
	char recieved(){
		return (!NewData_Recieved);
	}
	void togle_hnd_shk() {
		NewData_Recieved = (!NewData_Recieved);
	}
	int rem_lngth() {
		return input_buffer->get_length();
	}
	char dec_rem_lngth() {
		return input_buffer->dec_length();
	}
	void delete_buffer() {
		delete input_buffer;
		input_buffer = NULL;
	}
	buffer* get_buffer() {
		return input_buffer;
	}
};


class p_channels {
private:
	v_channels* vir_chanl;  //[N_channels];
	int vc_on_pc;
	buffer* waiting_channels;
	long int to_node;
	int dim_dir_chnl;
	static long int finished_messages;
	static unsigned long acum_exec_time;
public:
	char empty;
	p_channels** from_channels;
	p_channels* from_injection_channel;
	p_channels(int in_dim_dir_chnl);
	static unsigned long get_acum_exec_time () {
		return acum_exec_time;
	}
	static void reset_acum_exec_time () {
		acum_exec_time = 0;
	}

	void del_all() {
		buffer* temp;
		empty = 1;
		while (waiting_channels != NULL) {
			temp = waiting_channels;
			waiting_channels = waiting_channels->next;
			delete temp;
		}
		for (int i=0; i<N_channels; i++)
			if (vir_chanl[i].alocated()) {
				if (to_node == -1)
					vir_chanl[i].get_buffer()->del_destination_address();
				vir_chanl[i].delete_buffer();
				if (!vir_chanl[i].recieved())
					vir_chanl[i].togle_hnd_shk();
			}
	}
	buffer* rout(int*, unsigned long int*, int, int, int, int*, int, v_channels*, buffer** possible, 
		      int possible_ptr, char, char);
	char switch_channel(Time*, int&, int&);
	v_channels* virtual_channel(int channel) {
		return &vir_chanl[channel];
	}
	static long int get_finished_messages() {
	return finished_messages;
	}
	static void reset_finished_messages() {
	finished_messages = 0;
	}
	void set_to_node(long int in_to_node) {
		to_node = in_to_node;
	}
	long int get_to_node() {
		return to_node;
	}
	int get_dim_dir_chnl(){
		return dim_dir_chnl;
	}
};

class node {
private:
	p_channels** input_channels;
	p_channels** output_channels;
	int* node_address;
	p_channels* injection;
	p_channels* ejection;
public:
	void del_all () {
		for (int i = 0; i<N_dimensns; i++)
			output_channels[i]->del_all();
		injection->del_all();
		ejection->del_all();
	}
	void inject(int*, unsigned long int*, int);
	void connect_to(int, p_channels*);
	void connect_from(int, p_channels*);
	void set_node_address(long int);
	void connect_channels(int*, unsigned long int*, int, int, int, int*, v_channels*);
	node() {
		input_channels = new p_channels* [2 * N_dimensns];
		output_channels = new p_channels* [2 * N_dimensns];
		node_address =  new int [N_dimensns];
		injection = new p_channels(-1);
		ejection = new p_channels(-1);
		ejection->from_channels = input_channels;
		ejection->from_injection_channel = injection;
		ejection->set_to_node(-1);
	}
	void init_injection_channel(long int node_num) {
		injection->set_to_node(node_num);
	}
};

class network {
private:
	node* net_node;
	unsigned long int last_batch_time;
public:
	network();
	void del_all() {
		for (long int i = 0; i<N_nodes; i++)
			net_node[i].del_all();
	}
	long int static get_node(int[]);
	void simulate(Time*);
	void inject_message(long int, long int, int);
};

class event{
protected:
	int operation;
	event* other_event;
public:
	void set_others(event* in_other_event) {
		other_event = in_other_event;
	}
	event * next_event(){
		return other_event;
	}
	int operat(){
		return operation;
	}
};

class hnd_shk_event : event {
private:
	int vc_chnl;
	p_channels* on_physical;
public:
	hnd_shk_event(p_channels*, int);
	p_channels* phys_c() {
		return on_physical;
	}
	int vc_channel() {
		return vc_chnl;
	}
};

class inject_msg_event : event {
private:
	long int injecedt2node;
	long int destntn_node;
	int message_length;

public:
	inject_msg_event(long int, long int, int);
	int long injctd2n(){
		return injecedt2node;
	}
	int long dest_n(){
		return destntn_node;
	}
	int m_length(){
		return message_length;
	}
};

class rout_header_event : event {
private:
	buffer* last_buffer;
	long int at_node;
	int dim_dir_chnl;
	int v_chanl;
	v_channels* from_v_channel;
public:
	v_channels* get_from_v_channel () {
		return from_v_channel;
	}
	long int get_at_node() {
		return at_node;
	}
	int get_dim_dir_chnl() {
		return dim_dir_chnl;
	}
	int get_v_chanl() {
		return v_chanl;
	}
	rout_header_event(long int, buffer*, int, int, v_channels*);

	buffer* get_buffer_last() {
		return last_buffer;
	}
};

struct chnl_2b_switched {
	chnl_2b_switched* next;
	p_channels* channel;
};

class Time_instance {
public:
	unsigned long int Time_stamp;
	chnl_2b_switched* current_2b_switched;
	void del_all() {
		chnl_2b_switched* temp;
		while (current_2b_switched != NULL) {
			temp = current_2b_switched;
			current_2b_switched = current_2b_switched->next;
			delete temp;
		}
		event* temp_e;
		while (current_events != NULL) {
			temp_e = current_events;
			current_events = current_events->next_event();
			delete temp_e;
		}
	}
	event* current_events;
	Time_instance() {
		current_2b_switched = NULL;
	}
};

class Time {
private:
	Time_instance Time_array[MAX_Time];
public:
	long int current_Time;
	void del_all() {
		for (long int i=0; i<MAX_Time; i++)
			Time_array[i].del_all();
	}
	void insert_event(long int clock_from_now,
					event* in_event);
	event* event_2b_executed();
	Time();
	void add_next_2b_switched(int clock_from_now) {
		chnl_2b_switched* temp = Time_array[current_Time].current_2b_switched;
		Time_array[current_Time].current_2b_switched = temp->next;
		temp->next = Time_array[(current_Time+clock_from_now)%MAX_Time].current_2b_switched;
		Time_array[(current_Time+clock_from_now)%MAX_Time].current_2b_switched = temp;
	}
	void del_crnt_2b_switched() {
		chnl_2b_switched* temp = Time_array[current_Time].current_2b_switched;
		if (temp != NULL) {
			Time_array[current_Time].current_2b_switched =
				Time_array[current_Time].current_2b_switched->next;
			delete temp;
		}
	}
	p_channels* next_b_switched() {
		if (Time_array[current_Time].current_2b_switched == NULL)
			return NULL;
		return Time_array[current_Time].current_2b_switched->channel;
	}
	void add_2b_switched(int time_from_now, p_channels* chnl_2b_swtchd) {
		chnl_2b_switched* temp_2b_switched = new chnl_2b_switched;
		temp_2b_switched->channel = chnl_2b_swtchd;
		temp_2b_switched->next = Time_array[(current_Time+time_from_now)%MAX_Time].current_2b_switched;
		Time_array[(current_Time+time_from_now)%MAX_Time].current_2b_switched = temp_2b_switched;
	}
	unsigned long int get_current_time() {
		return current_Time;
	}
};

/////////////////////////////////////////////////////////////////////////////

	buffer::buffer(int* dst_adrs, unsigned long int* in_insrtn_time_stamp, int lngth, 
			int in_physical_c, int in_virtual_c, int* in_source, int in_frwd_bkwd, 
			v_channels* in_from_v_channel, p_channels* in_channel_ahead, 
			buffer** in_possible, int in_possible_ptr, buffer* in_last_buf, 
			char in_definit, buffer** in_waiting, char in_deterministic){
		destination_address = dst_adrs;
		length = lngth;
		virtual_c = in_virtual_c;
		physical_c = in_physical_c;
		next = NULL;
		source = in_source;
		frwd_bkwd = in_frwd_bkwd;
		from_channel = in_from_v_channel;
		insrtn_time_stamp = in_insrtn_time_stamp;
		channel_ahead = in_channel_ahead;
		possible = in_possible;
		possible_ptr = in_possible_ptr;
		last_buf = in_last_buf;
		definit = in_definit;
		waiting_c = in_waiting;
		deterministic = in_deterministic;
	}

	char buffer::last_new_data() {
		if (from_channel == NULL) return 1;
		return !from_channel->recieved();
	}


///////////////////////////////////////////////////////////////////////////////////////////

	v_channels::v_channels(){
		NewData_Recieved = 0;
		input_buffer = NULL;
	}

	char v_channels::alocated(){
		if (input_buffer == NULL)
			return 0;
		else return 1;
	}

	void v_channels::aloct(buffer* input) {
		input_buffer = input;             // 'NewData' signal activated by event in next clock
	}

///////////////////////////////////////////////////////////////////////////////////////////

long int p_channels::finished_messages = 0;
unsigned long p_channels::acum_exec_time = 0;

	p_channels::p_channels(int in_dim_dir_chnl) {
		vir_chanl = new v_channels [N_channels];
		waiting_channels = NULL;
		vc_on_pc = 0;
		dim_dir_chnl = in_dim_dir_chnl;
		empty = 1;
	};

	char p_channels::switch_channel(Time* main_Time, int& from_pc, int& from_vc){
		int last_vc_on_pc = vc_on_pc;
		vc_on_pc = (vc_on_pc + 1)%N_channels;

again:	if (waiting_channels != NULL)
			if (!vir_chanl[vc_on_pc].alocated()){
				if ((vc_on_pc != 0) || waiting_channels->deterministic)
				{
					for (int i = 0; i < waiting_channels->possible_ptr; i++)
						if (waiting_channels != waiting_channels->possible[i]) {
							waiting_channels->possible[i]->adjust_ptrs();
							delete waiting_channels->possible[i];
							waiting_channels->possible[i] = NULL;
						}
					delete[] waiting_channels->possible;
					waiting_channels->possible = NULL;
					waiting_channels->possible_ptr = 0;

					if (waiting_channels->from_vc() != -1)
						if (waiting_channels->from_pc() != -1) {
							if (from_channels[waiting_channels->from_pc()]->virtual_channel(waiting_channels->from_vc())->get_buffer()->channel_ahead == NULL)
								from_channels[waiting_channels->from_pc()]->virtual_channel(waiting_channels->from_vc())->get_buffer()->channel_ahead = this;
						} else
							if (from_injection_channel->virtual_channel(waiting_channels->from_vc())->get_buffer()->channel_ahead == NULL)
								from_injection_channel->virtual_channel(waiting_channels->from_vc())->get_buffer()->channel_ahead = this;

					vir_chanl[vc_on_pc].aloct(waiting_channels);
					from_pc = waiting_channels->from_pc();
					from_vc = waiting_channels->from_vc();
					waiting_channels =
						waiting_channels->next;
					if (waiting_channels != NULL)
						waiting_channels->last_buf = NULL;
					event* new_event =
						(event*) new hnd_shk_event(this, vc_on_pc);
					main_Time->insert_event(NO_CYCLES, new_event);
					if (get_to_node() != -1) {
						new_event = (event*) new rout_header_event(get_to_node(),
									vir_chanl[vc_on_pc].get_buffer(),
									get_dim_dir_chnl(),
									vc_on_pc,
									&vir_chanl[vc_on_pc]);
						main_Time->insert_event(NO_CYCLES, new_event);
					}
					return 1;
				}
				else {
					buffer* temp = waiting_channels;
					if (vc_on_pc == 0)
						while ((temp->next != NULL) && (!temp->next->deterministic))
							temp = temp->next;
					if (temp->next != NULL) {
						for (int i=0; i<temp->next->possible_ptr; i++)
							if (temp->next != temp->next->possible[i]) {
								temp->next->possible[i]->adjust_ptrs();
								delete temp->next->possible[i];
								temp->next->possible[i] = NULL;
							}
						delete[] temp->next->possible;
						temp->next->possible = NULL;
						temp->next->possible_ptr = 0;

						if (temp->next->from_vc() != -1)
							if (temp->next->from_pc() != -1) {
								if (from_channels[temp->next->from_pc()]->virtual_channel(temp->next->from_vc())->get_buffer()->channel_ahead == NULL)
									from_channels[temp->next->from_pc()]->virtual_channel(temp->next->from_vc())->get_buffer()->channel_ahead = this;
							} else
								if (from_injection_channel->virtual_channel(temp->next->from_vc())->get_buffer()->channel_ahead == NULL)
									from_injection_channel->virtual_channel(temp->next->from_vc())->get_buffer()->channel_ahead = this;

						vir_chanl[vc_on_pc].aloct(temp->next);
						from_pc = temp->next->from_pc();
						from_vc = temp->next->from_vc();
						temp->next = temp->next->next;
						if (temp->next != NULL)
							temp->next->last_buf = temp;
						event* new_event =
							(event*) new hnd_shk_event(this, vc_on_pc);
						main_Time->insert_event(NO_CYCLES, new_event);
						if (get_to_node() != -1) {
							new_event = (event*) new rout_header_event(get_to_node(),
										vir_chanl[vc_on_pc].get_buffer(),
										get_dim_dir_chnl(),
										vc_on_pc,
										&vir_chanl[vc_on_pc]);
							main_Time->insert_event(NO_CYCLES, new_event);
						}
						return 1;
					}
					else
						if (vc_on_pc != last_vc_on_pc) {
							vc_on_pc = (vc_on_pc+1)%N_channels;
							goto again;
						}
				}
			}

		while((!vir_chanl[vc_on_pc].alocated()) && (vc_on_pc != last_vc_on_pc))
			vc_on_pc = (vc_on_pc + 1)%N_channels;

		if (vir_chanl[vc_on_pc].alocated()) {
			if (vir_chanl[vc_on_pc].recieved() || (to_node == -1)) {
				if (vir_chanl[vc_on_pc].get_buffer()->last_new_data()
					|| (vir_chanl[vc_on_pc].get_buffer()->get_length() < 2))
				{
					if (vir_chanl[vc_on_pc].dec_rem_lngth()){
						hnd_shk_event* new_event = new hnd_shk_event(this, vc_on_pc);
						main_Time->insert_event(1, (event*)new_event);
						from_vc = (vir_chanl[vc_on_pc].get_buffer())->from_vc();
						from_pc = (vir_chanl[vc_on_pc].get_buffer())->from_pc();
						return 1;
					}
					if (to_node == -1){
						vir_chanl[vc_on_pc].get_buffer()->del_destination_address();
						finished_messages++;

						acum_exec_time += (MAX_Time + main_Time->get_current_time() - 
						*(vir_chanl[vc_on_pc].get_buffer()->get_insrtn_time_stamp()))
													%MAX_Time;
						delete vir_chanl[vc_on_pc].get_buffer()->get_insrtn_time_stamp();
					}
					vir_chanl[vc_on_pc].delete_buffer();
					return 2;
				}
			}
			if (vc_on_pc != last_vc_on_pc) {
				vc_on_pc = (vc_on_pc+1)%N_channels;
				goto again;
			}
		}
			empty = 1;
			return 0;
	}


	buffer* p_channels::rout(int* dst_adrs, unsigned long int* insertion_time_stamp, int lngth, 
					int in_physical_c, int in_virtual_c, int* source, int frwd_bkwd, 
					v_channels* in_from_v_channel, buffer** possible, int possible_ptr, 
					char definit, char in_deterministic){
		buffer** temp = &waiting_channels;
		buffer* prev_buf = NULL;
		while (*temp != NULL) {
			prev_buf = *temp;
			temp = &((*temp)->next);
		}
		*temp = new buffer(dst_adrs, insertion_time_stamp, lngth, in_physical_c, in_virtual_c, 
					source, frwd_bkwd, in_from_v_channel, NULL, possible, 
					possible_ptr, prev_buf, definit, &waiting_channels, in_deterministic);
		if (empty) {
			empty = 0;
			main_Time->add_2b_switched(0, this);
		}
		if (in_from_v_channel != NULL)
			if (definit)
				in_from_v_channel->get_buffer()->channel_ahead = this;
			else in_from_v_channel->get_buffer()->channel_ahead = NULL;
		return *temp;
	}

//////////////////////////////////////////////////////////////////////////////////////////

	void node::inject(int* dst_adrs,unsigned long int* insertion_time_stamp, int lngth) {
		injection->rout(dst_adrs, insertion_time_stamp, lngth, -1, -1, node_address, 0, NULL, NULL, NULL, 1, 1);
	}

	void node::connect_channels(int* dst_adrs, unsigned long int* in_insertion_time_stamp, int lngth, 
					int in_physical_c, int in_virtual_c, int* source, 
					v_channels* in_from_v_channel ){
		int temp = 0;
		int temp1;
		char found_channel = 0;
		int fwrd_bwrd;
		int possible_dimensions_no[N_dimensns];
		int possible_fwrd_bwrd[N_dimensns];
		int possible_dimensions_ptr = 0;
		while (temp < N_dimensns) {
			if ( dst_adrs[temp] != node_address[temp] ) {
				found_channel = 1;
				if (!wrap_around) {
					if (dst_adrs[temp]>node_address[temp])
						temp1 = 2*temp;
					else
						temp1 = 2*temp+1; 
				}
				else {
					if (
					     (
					      ((dst_adrs[temp]-node_address[temp]+L_dimensns)%L_dimensns) <
					      ((node_address[temp]-dst_adrs[temp]+L_dimensns)%L_dimensns)  
					     )
					   )
						temp1 = 2*temp;
					else
						temp1 = 2*temp+1; 
				}								
				possible_dimensions_no[possible_dimensions_ptr] = temp1;
				possible_fwrd_bwrd[possible_dimensions_ptr++] = dst_adrs[temp] - source[temp];
				if (!adaptive)
					temp = N_dimensns;   // dimension order
			}
			temp++;
		}

		if (found_channel) {
			buffer** possible_dimensions = new buffer* [N_dimensns];
			for (int i=0; i < possible_dimensions_ptr; i++)
				possible_dimensions[i] =
					output_channels[possible_dimensions_no[i]]->
					rout(dst_adrs, in_insertion_time_stamp, lngth, in_physical_c, 
						in_virtual_c, source, possible_fwrd_bwrd[i], 
						in_from_v_channel, possible_dimensions, 
						possible_dimensions_ptr, 0, !i);
		return;
		}

		ejection->rout(dst_adrs, in_insertion_time_stamp, lngth, in_physical_c, 
				in_virtual_c, source, 0, in_from_v_channel, NULL, 0, 1, 1);

	}

	void node::connect_to(int dim, p_channels* out_channel) {
		output_channels[dim] = out_channel;
		output_channels[dim]->from_channels = input_channels;
		output_channels[dim]->from_injection_channel = injection;
	}

	void node::connect_from(int dim, p_channels* in_channel) {
		input_channels[dim] = in_channel;
	}

	void node::set_node_address(long int i) {
		long int temp = i;
		for (int j=0; j<N_dimensns; j++){
			node_address[j] = temp%L_dimensns;
			temp /= L_dimensns;
		}
	}

////////////////////////////////////////////////////////////////////////////////////////

	network::network(){
		net_node = new node [N_nodes];
		for (long int i=0; i<N_nodes; i++) {
			net_node[i].set_node_address(i);
			net_node[i].init_injection_channel(i);
			last_batch_time = 0;
			}

		for (long int i=0; i< N_nodes; i++){
			long int temp = i;
			int node_address[N_dimensns];
			for	(int j=0; j<N_dimensns; j++){
				node_address[j] = temp%L_dimensns;
				temp /= L_dimensns;
			}
			for (int j = 0; j<N_dimensns; j++){
				int temp = node_address[j];
				node_address[j] = (temp+1)%L_dimensns;
				p_channels* temp_ptr = new p_channels(2*j);
				net_node[i].connect_to(2*j, temp_ptr);
				net_node[get_node(node_address)].connect_from(2*j, temp_ptr);
				temp_ptr->set_to_node(get_node(node_address));
				node_address[j] = (temp-1)%L_dimensns;
				if (node_address[j]<0)
					node_address[j] += L_dimensns;
				temp_ptr = new p_channels(2*j+1);
				net_node[i].connect_to(2*j+1, temp_ptr);
				net_node[get_node(node_address)].connect_from(2*j+1, temp_ptr);
				temp_ptr->set_to_node(get_node(node_address));
				node_address[j] = temp;
			}
		}
	}

	long int network::get_node(int node_address[]) {
		long int temp = 0;
		long int radix = 1;
		for (int i=0; i<N_dimensns; i++){
			temp += node_address[i]*radix;
			radix *= L_dimensns;
		}
		return temp;
	}

	void network::simulate(Time* main_Time) {

		long int execution_time;
		outp_file =  fopen(output, "w");
		if(outp_file==NULL) {
			printf("File error\n");
			return;
		}

		srand((unsigned)time( NULL ));
		srand(rand());
			for (long int i=0; i<N_nodes; i++) {
				long int test1 = rand() % (N_nodes);
				long int test2 = rand() % (N_nodes);
				inject_msg_event* first_event = new inject_msg_event (i, test1, MESSAGE_LENGTH);
				main_Time->insert_event(0, (event *)first_event);
			}

		global1 = 0;
		for (double gen_rate = gen_period; gen_rate <= stop_period; gen_rate += period_step ) {
			GEN_PRD = int (1.0/gen_rate);
			printf("\ngeneration period = %d\n", GEN_PRD);
//			fprintf(outp_file, "\ngeneration period = %d\n", GEN_PRD);
			batches_counted = 0;
			double acum_batchs=0;
			while (batches_counted<22){
				event* curnt_event;
				while ((curnt_event = main_Time->event_2b_executed()) != NULL) {
					if (curnt_event->operat() == INJECT) {
						inject_msg_event* this_event = (inject_msg_event*) curnt_event;
						long int temp_node_num = this_event->dest_n();
						unsigned long int* insertion_time_stamp = new unsigned long int;
						*insertion_time_stamp = main_Time->get_current_time();
						int* node_address = new int[N_dimensns];
						for	(int j=0; j<N_dimensns; j++) {
							node_address[j] = temp_node_num%L_dimensns;
							temp_node_num /= L_dimensns;
						}

						net_node[this_event->injctd2n()].inject(node_address, 
											insertion_time_stamp, 
											this_event->m_length());
						global1++;

						long int test1 = rand() % (N_nodes);
						inject_msg_event* first_event = 
							new inject_msg_event (this_event->injctd2n(), test1, 
												MESSAGE_LENGTH);
						long int next_time = 
							log((double)RAND_MAX/((rand())+1)) * GEN_PRD * NO_CYCLES;
						main_Time->insert_event(next_time, (event *)first_event);
						delete this_event;
					}
					else if (curnt_event->operat() == HND_SHK) {
						hnd_shk_event* this_event = (hnd_shk_event*) curnt_event;
						v_channels* temp_vc;
						int test = this_event->vc_channel();
						temp_vc = (this_event->phys_c())->virtual_channel(this_event->vc_channel());
						temp_vc->togle_hnd_shk();
						 if (this_event->phys_c()->empty) {
							this_event->phys_c()->empty = 0;
							main_Time->add_2b_switched(0, this_event->phys_c());
						 }
						 if (temp_vc->get_buffer()->channel_ahead != NULL)
							 if (temp_vc->get_buffer()->channel_ahead->empty) {
								temp_vc->get_buffer()->channel_ahead->empty = 0;
								main_Time->add_2b_switched(0, 
									temp_vc->get_buffer()->channel_ahead);
							 }
						delete this_event;
					}
					else if (curnt_event->operat() == ROUT_HEADER) {
						rout_header_event* this_event = (rout_header_event*) curnt_event;
						buffer* last_buffer = this_event->get_buffer_last();
						int* dest_adrs = last_buffer->get_destination_address();
						int length = last_buffer->get_length();
						unsigned long int * insertion_time_stamp = 
								last_buffer->get_insrtn_time_stamp();
						net_node[this_event->get_at_node()].
							connect_channels(dest_adrs, insertion_time_stamp, 
									   length, this_event->get_dim_dir_chnl(), 
									   this_event->get_v_chanl(), 
									   last_buffer->get_source(), 
									   this_event->get_from_v_channel());
						delete this_event;
					}
				}

				while (main_Time->next_b_switched() != NULL) {
					int from_pc, from_vc;
					char temp_swtch_rtn;
					if ((temp_swtch_rtn = 
						main_Time->next_b_switched()->switch_channel(main_Time, 
												from_pc, from_vc)) 
													== 1) {
						hnd_shk_event* new_event;
						if (from_vc != -1) {
							if (from_pc != -1)
								new_event = new hnd_shk_event(main_Time->next_b_switched()->from_channels[from_pc], from_vc);
							else
								new_event = new hnd_shk_event(main_Time->next_b_switched()->from_injection_channel, from_vc);
							main_Time->insert_event(1, (event*)new_event);
						}
					}
					if (temp_swtch_rtn)
						main_Time->add_next_2b_switched(NO_CYCLES);
					else main_Time->del_crnt_2b_switched();
				}

				main_Time->current_Time = (main_Time->current_Time+1)%MAX_Time;

				if (p_channels::get_finished_messages() > BATCH_SIZE){
					batches_counted++;
					execution_time = (MAX_Time + main_Time->get_current_time() 
										- last_batch_time)%MAX_Time;
					last_batch_time = main_Time->get_current_time();
					printf("%d         %f\n", execution_time,
								((double)p_channels::get_acum_exec_time()
									/p_channels::get_finished_messages())
												/NO_CYCLES );
/*					fprintf(outp_file, "%d         %f\n", execution_time,
								((double)p_channels::get_acum_exec_time()
									/p_channels::get_finished_messages())
												/NO_CYCLES );*/
					if (batches_counted > 6) acum_batchs += 
								((double)p_channels::get_acum_exec_time()
									/p_channels::get_finished_messages())
									/NO_CYCLES;
					p_channels::reset_finished_messages();
					p_channels::reset_acum_exec_time();
					fflush(outp_file);
				}
			}
			printf("\nAverage: %f\n", acum_batchs/16);
			fprintf(outp_file, "%f\t%f\n",gen_rate, acum_batchs/16);
		}
	fclose(outp_file);
	}


/////////////////////////////////////////////////////////////////////////////////

inject_msg_event::inject_msg_event(long int in_injecedt2node, long int in_destntn_node, int in_message_length) {
	operation = INJECT;
	injecedt2node = in_injecedt2node;
	destntn_node = in_destntn_node;
	message_length = in_message_length;
}

hnd_shk_event::hnd_shk_event(p_channels* physical_pntr, int in_vc_chnl) {
	operation = HND_SHK;
	on_physical = physical_pntr;
	vc_chnl = in_vc_chnl;
}

rout_header_event::rout_header_event(long int in_at_node, buffer* in_last_buffer,
					int in_dim_dir_chnl, int in_v_chanl, v_channels* in_from_v_channel) {
		operation = ROUT_HEADER;
		at_node = in_at_node;
		dim_dir_chnl = in_dim_dir_chnl;
		v_chanl = in_v_chanl;
		last_buffer = in_last_buffer;
		from_v_channel = in_from_v_channel;
	}

Time::Time() {
	current_Time = 0;
	for (long int i=0; i<MAX_Time; i++){
		Time_array[i].Time_stamp = i;
		Time_array[i].current_events = NULL;
	}
}

void Time::insert_event(long int clocks_from_now,
					event* in_event) {
	in_event->set_others(Time_array[(clocks_from_now+current_Time)%MAX_Time].current_events);
	Time_array[(clocks_from_now+current_Time)%MAX_Time].current_events = in_event;
}

event* Time::event_2b_executed() {
	event* temp;
	temp = Time_array[current_Time].current_events;
	if (temp != NULL)
		Time_array[current_Time].current_events = temp->next_event();
	return temp;
}

////////////////////////////////////////////////////////////////////////////////////////
int main (int argc, char **argv) {
	int starting_index = 0;
	adaptive = 0;
	wrap_around = 0;
	for (int i=0; i<argc; i++) {
		if (!strcmp("-a", argv[i])) {
			adaptive = 1;
			starting_index++;
		} else if (!strcmp("-w", argv[i])) {
			wrap_around = 1;
			starting_index++;
		}
	}
				 
	if ((argc - starting_index != 9 ))  {
		printf("\nFORMAT: [-flags] <number of dimensions> <radix>");
		printf("\n                 <number of channels> <message length>");
		printf("\n                 <initial gen. freq> <gen. freq steps> <last gen. freq>");
		printf("\n                 <output file>\n"); 
		printf("\nFLAGS:"); 
		printf("\n       -a : adaptive routing [default is dimension-order]"); 
		printf("\n       -w : use wraparound channels [default is a mesh network]\n\n");
		return 0;
	}
	
	N_nodes = 1;
	for (int i=0; i < atoi(argv[1 + starting_index]); i++)
		N_nodes *= atoi(argv[2 + starting_index]);
	N_dimensns = atoi(argv[1 + starting_index]);
	L_dimensns = atoi(argv[2 + starting_index]);
	N_channels = atoi(argv[3 + starting_index]);
	MESSAGE_LENGTH = atoi(argv[4 + starting_index]);
	gen_period = atof(argv[5 + starting_index]);
	period_step = atof(argv[6 + starting_index]);
	stop_period = atof(argv[7 + starting_index]);
	output = argv[8 + starting_index];

	main_Time = new Time;
	network main;
	main.simulate(main_Time);
}
